use std::{ fs, io::{Bytes, stdin, BufReader, Error, ErrorKind, Read, Write}, net::{IpAddr, Ipv4Addr, SocketAddrV4, TcpStream}, path::Path, thread::{self, JoinHandle}, time::Duration};
use chrono::{DateTime, Local};
use image::{Rgb, RgbImage};
use ipnet::{Ipv4Net,PrefixLenError};
use local_ip_address::linux::local_ip;
use log::{LevelFilter, trace,debug,warn,error,info};
use fern::{log_file, Dispatch};

//RTSP port is 554; prepare for real world situations
const PORT_NUMBER:u16 = 55001;
const PORT_PROBE_TIMEOUT:Duration = Duration::new(5,0);
const SUBNET_SIZE:u8 = 24;

//These subnets are well documented;
///RIT owns the subnet 129.21.0.0/16; probing by attempting to open thousands of TCP streams
///simultaneously is unwise.
const UNSAFE_SUBNET:Result<Ipv4Net,PrefixLenError> = Ipv4Net::new(Ipv4Addr::new(129,21,0,0),16);

///This subnet is universally considered the "local" subnet. This results in 256 addresses, and can
///be probed both safely and efficiently, and as long as the world server on the LAN is running it 
///is functionally guaranteeed to be found.
const HOME_SUBNET:Result<Ipv4Net,PrefixLenError> = Ipv4Net::new(Ipv4Addr::new(192,168,0,0),24);

fn main() {
    setup_logs(&true);
    //Get host IP
    let host_ip = local_ip().expect("No real IP address!");
    trace!("Host IP: {:?}",host_ip);

    let mut thread_list:Vec<JoinHandle<Result<TcpStream,Error>>> = Vec::new();
    let mut port_list:Vec<TcpStream> = Vec::new();
    if let IpAddr::V4(host_v4_ip) = host_ip{
        //Check current subnet
        let net = Ipv4Net::new(host_v4_ip,SUBNET_SIZE);

        if let Ok(net) = net{
            //RIT owns 129.21.0.0/16; DO NOT SPAM THIS NETWORK
            if UNSAFE_SUBNET.unwrap().contains(&net){
                //Request world server address from user
                let address = manual_address_prompt();
                if address.is_ok(){
                    port_list.push(address.unwrap());
                    //thread_list won't be added to, so the thread for loop won't execute
                }
            } else if net.contains(&HOME_SUBNET.unwrap()){
                //Collect all Hosts on the current subnet
                info!("aggregating all IPs... This may take several minutes...");
                for net_address in net.hosts(){
                    thread_list.push(
                        thread::spawn(move ||{
                            let ip_port = std::net::SocketAddr::V4(SocketAddrV4::new(net_address, PORT_NUMBER));
                            //Attempt to open a TCP stream with this IP address, exit the thread
                            //returning the open TcpStream, or an error
                            return TcpStream::connect_timeout(&ip_port, PORT_PROBE_TIMEOUT);
                        })
                    );
                }
            }
        }
    }

    //Multithread the port-sniffing block, only save sucessful connections
    for thread in thread_list{
        let output = thread.join().unwrap();
        if let Ok(real_output) = output{
            port_list.push(real_output);
        }
    }

    //For all available ports, try the custom communication protocol
    for mut stream in port_list{
        info!("Stream connected to address: {:?}",stream.peer_addr());
        send_processing_time(&mut stream, 2.7f64);
        read_response(&mut stream);
    }
    //TODO: Document/implement RTSP client-side comms
    //Currently using custom port and custom protocol, but if real-world cameras will 
    //be in use, we will likely need to be portable to the RTSP protocol, or similar
}

//Used to send the calculated time value to the world server
pub fn send_processing_time(stream:&mut TcpStream ,sent_value:f64){
        debug!("Begin handshake; sending timestamp float...");
        //Write the f64 value, in bytes, to the stream, and ensure it has been sent
        stream.write(&sent_value.to_le_bytes()).unwrap();
        stream.flush().unwrap();
        debug!("Timestamp sent!");
}

//Used to recieve images from the world server
pub fn read_response(stream:&mut TcpStream){
    //Buffer the incoming stream
    let reader = BufReader::new(stream.try_clone().expect("Failed clone"));
    //Create a storage point for the bytes
    let mut recieved_bytes:Vec<u8> = Vec::new();
    //Create an iterator over the stream's bytes
    let mut iter = reader.bytes();

    debug!("Continue handshake, read back image packet length...");
    //The first values sent will always be a u32 containing the length of the message (without the
    //length counted), followed by the image width and height
    let length = bytes_to_u32(&mut iter);
    debug!("Image packet length read as {:?}!",length);

    debug!("Continue handshake, read back image width...");
    let image_width = bytes_to_u32(&mut iter);
    debug!("Image width read as {:?}!",image_width);

    debug!("Continue handshake, read back image height...");
    let image_height = bytes_to_u32(&mut iter);
    debug!("Image height read as {:?}!",image_height);

    //The length of the image is the total number of bytes
    let byte_depth = length / (image_width * image_height);
    debug!("Image bitdepth calculated as {:?}!", byte_depth);

    //Confirm to world server the correct number of bytes
    stream.write(&byte_depth.to_le_bytes()).unwrap();
    stream.flush().unwrap();

    //TODO: Remove assumption
    //Always assumes world server will send image
    info!("Start reading...");
    //Recieve the rest of the message
    for _ in 0..length{
        recieved_bytes.push(iter.next().unwrap().unwrap_or(0));
    }

    info!("Image successfully recieved");

    //Create a blank pane on which to store the read in pixels
    let mut img = RgbImage::new(image_width,image_height);

    //Create a temporary storage point for u8s when generating pixels
    let mut temp:Vec<u8> = Vec::new();

    //Iterate over bytes
    info!("Building image...");
    for i in 0..recieved_bytes.len(){
        let byte = recieved_bytes.get(i).expect("Index does not exist!");

        temp.push(byte.clone());
        //Each pixel is composed of [byte_depth] u8s
        if ((i as u32)+1) % byte_depth == 0{
            //Coordinate in the pane = i/3 
            let pixel_number:u32 = (i as u32) /byte_depth;
            //Insert the pixel into the pane
            img.put_pixel(
                //Pixels coming directly from Unity are sent "upside down", that is,
                //starting at the bottom right, and iterating horizontally, then vertically.
                //This line compensates and creates a properly-oriented image
                (image_width-1)-(pixel_number%image_width), (image_height-1)-(pixel_number / image_width), 
                //Extract image from temporary storage, ang generate pixel
                Rgb::<u8>{0:temp.clone().try_into().expect("bad vector length!")}
            );
            //Clear temporary storage
            temp.clear();
        } 
    }
    //Dump image to file
    //ONLY FOR EXPERIMENTATION AND TESTING PURPOSES
    info!("Image built, saving to file...");
    _ = img.save("./test.png");
    info!("image saved!");
}

///Borrows an iterator over the bytes in a buffered-reader of a TcpStream; reads the next 4 bytes,
///outputs the generated u32
fn bytes_to_u32(iter:&mut Bytes<BufReader<TcpStream>>) -> u32{
    let mut array:[u8;4] = [0;4];
    for i in 0..4{
        array[i] = iter.next().unwrap().unwrap();
    }
    return u32::from_le_bytes(array);
}

///Request the user to input the world server's address into stdin
fn manual_address_prompt() -> Result<TcpStream,Error>{
    let return_val:Result<TcpStream,Error> = Err(Error::new(ErrorKind::AddrNotAvailable,"Invalid Address"));
    while return_val.is_err(){
        let mut user_input:String = String::default();
        match stdin().read_line(&mut user_input){
            Ok(_) => {
                match user_input.trim().parse::<SocketAddrV4>(){
                    Ok(address) => {
                        //Test if the address is accurate
                        let stream = TcpStream::connect_timeout(&std::net::SocketAddr::V4(address), PORT_PROBE_TIMEOUT);
                        if stream.is_ok(){
                            return stream;
                        } else {
                            warn!("Address unavailable!");
                            debug!("{:?}",stream.unwrap_err());
                        }
                    },
                    Err(error) => {
                        error!("Unable to properly parse user input!");
                        debug!("{:?}",error);
                    }
                }
            },
            Err(error) => {
                error!("Unable to read user input!");
                debug!("{:?}",error);
            }
        }
    }
    return return_val;
}

///Set up logging macros to be used throughout the program
fn setup_logs(debug:&bool){
    //Get the current time
    let chrono_now: DateTime<Local> = Local::now();
    //Create a log directory, if it does not exist
    if ! Path::new("logs").is_dir(){
        _ = fs::create_dir("logs");
    };
    //Create a log macro listener
    _ = fern::Dispatch::new()
        //Format the output messages
        .format(|out,message,record|{
            out.finish(format_args!(
                "{} - [{}, {}] - {}",
                Local::now().to_rfc3339(),
                record.level(),
                record.target(),
                message
            ))
        })
        //Output to the file
        .chain({
            let mut file_logger = Dispatch::new();
            let date_format = chrono_now.format("%Y-%m-%d_%H.%M").to_string();
            let local_log_file = log_file(format!("logs/{}.log",date_format)).unwrap();
            //Set filter based on whether the command is being run in "debug" mode
            if *debug{
                file_logger = file_logger.level(LevelFilter::Trace);
            }
            else {
                file_logger = file_logger.level(LevelFilter::Debug);
            }
            //Apply the filter and file formatting rules described above
            file_logger.chain(local_log_file)
        })
        .chain({
            //Also log to stdout
            let mut stdout_logger = fern::Dispatch::new();
            if *debug {
                stdout_logger = stdout_logger.level(LevelFilter::Debug);
            }
            else {
                stdout_logger = stdout_logger.level(LevelFilter::Info);
            }
                stdout_logger.chain(std::io::stdout())
        })
        //Start listening to macros
        .apply();
}
